@@ -38,7 +38,7 @@
 #include <unistd.h>
 #include <time.h>
 #include <string.h>
-#include <libgnome/gnome-i18n.h>
+#include <glib/gi18n.h>
 #include <libgnome/gnome-util.h>
 #include <libgnome/gnome-exec.h>
 #include <libgnome/gnome-url.h>
@@ -50,7 +50,6 @@
 #include <libgnomevfs/gnome-vfs-utils.h>
 
 #include <libgnome/gnome-desktop-item.h>
-#include "gnome-desktop-i18n.h"
 
 #ifdef HAVE_STARTUP_NOTIFICATION
 #define SN_API_NOT_YET_FROZEN
@@ -88,6 +87,8 @@ struct _GnomeDesktopItem {
 	char *location;
 
 	time_t mtime;
+
+	guint32 launch_time;
 };
 
 /* If mtime is set to this, set_location won't update mtime,
@@ -367,6 +368,8 @@ gnome_desktop_item_new (void)
 				       GNOME_DESKTOP_ITEM_VERSION,
 				       "1.0");
 
+	retval->launch_time = 0;
+
 	return retval;
 }
 
@@ -419,6 +422,7 @@ gnome_desktop_item_copy (const GnomeDesktopItem *item)
 	retval->modified = item->modified;
 	retval->location = g_strdup (item->location);
 	retval->mtime = item->mtime;
+	retval->launch_time = item->launch_time;
 
 	/* Languages */
 	retval->languages = g_list_copy (item->languages);
@@ -914,16 +918,16 @@ lookup_locale (const GnomeDesktopItem *item, const char *key, const char *locale
 static const char *
 lookup_best_locale (const GnomeDesktopItem *item, const char *key)
 {
-	const GList *list = gnome_i18n_get_language_list ("LC_MESSAGES");
-	while (list != NULL) {
-		const char *ret;
-		const char *locale = list->data;
+	const char * const *langs_pointer;
+	int                 i;
 
-		ret = lookup_locale (item, key, locale);
+	langs_pointer = g_get_language_names ();
+	for (i = 0; langs_pointer[i] != NULL; i++) {
+		const char *ret = NULL;
+
+		ret = lookup_locale (item, key, langs_pointer[i]);
 		if (ret != NULL)
 			return ret;
-
-		list = list->next;
 	}
 
 	return NULL;
@@ -1024,7 +1028,8 @@ make_args (GList *files)
 		if (file == NULL)
 			continue;;
 		uri = gnome_vfs_uri_new (file);
-		list = g_slist_prepend (list, uri);
+		if (uri)
+			list = g_slist_prepend (list, uri);
 	}
 
 	return g_slist_reverse (list);
@@ -1790,6 +1795,7 @@ ditem_execute (const GnomeDesktopItem *item,
 #ifdef HAVE_STARTUP_NOTIFICATION
 		if (sn_context != NULL &&
 		    !sn_launcher_context_get_initiated (sn_context)) {
+			guint32 launch_time;
 
 			/* This means that we always use the first real_argv[0]
 			 * we select for the "binary name", but it's probably
@@ -1801,10 +1807,18 @@ ditem_execute (const GnomeDesktopItem *item,
 			sn_launcher_context_set_binary_name (sn_context,
 							     real_argv[0]);
 			
+			if (item->launch_time > 0)
+				launch_time = item->launch_time;
+			else
+				launch_time = gtk_get_current_event_time ();
+
 			sn_launcher_context_initiate (sn_context,
 						      g_get_prgname () ? g_get_prgname () : "unknown",
 						      real_argv[0],
-						      gtk_get_current_event_time ());
+						      launch_time);
+
+			/* Don't allow accidental reuse of same timestamp */
+			((GnomeDesktopItem *)item)->launch_time = 0;
 
 			envp = make_spawn_environment_for_sn_context (sn_context, envp);
 			if (free_me)
@@ -2807,16 +2821,16 @@ const char *
 gnome_desktop_item_get_attr_locale (const GnomeDesktopItem *item,
 				    const char             *attr)
 {
-	const GList *list, *l;
+	const char * const *langs_pointer;
+	int                 i;
 
-	list = gnome_i18n_get_language_list ("LC_MESSAGES");
-	for (l = list; l; l = l->next) {
-		const char *locale = l->data;
+	langs_pointer = g_get_language_names ();
+	for (i = 0; langs_pointer[i] != NULL; i++) {
 		const char *value = NULL;
 
-		value = lookup_locale (item, attr, locale);
+		value = lookup_locale (item, attr, langs_pointer[i]);
 		if (value)
-			return locale;
+			return langs_pointer[i];
 	}
 
 	return NULL;
@@ -2846,14 +2860,15 @@ gnome_desktop_item_get_languages (const GnomeDesktopItem *item,
 static const char *
 get_language (void)
 {
-	const GList *list = gnome_i18n_get_language_list ("LC_MESSAGES");
-	while (list != NULL) {
-		const char *lang = list->data;
+	const char * const *langs_pointer;
+	int                 i;
+
+	langs_pointer = g_get_language_names ();
+	for (i = 0; langs_pointer[i] != NULL; i++) {
 		/* find first without encoding  */
-		if (strchr (lang, '.') == NULL) {
-			return lang; 
+		if (strchr (langs_pointer[i], '.') == NULL) {
+			return langs_pointer[i]; 
 		}
-		list = list->next;
 	}
 	return NULL;
 }
@@ -2976,8 +2991,18 @@ gnome_desktop_item_set_boolean (GnomeDesktopItem *item,
 	set (item, attr, value ? "true" : "false");
 }
 
+void
+gnome_desktop_item_set_launch_time (GnomeDesktopItem *item,
+				    guint32           timestamp)
+{
+	g_return_if_fail (item != NULL);
+	g_return_if_fail (timestamp > 0);
+
+	item->launch_time = timestamp;
+}
+
 /*
-* Clearing attributes
+ * Clearing attributes
  */
 void
 gnome_desktop_item_clear_section (GnomeDesktopItem *item,
